(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2014 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

BeginPackage["AtomicDensityMatrix`ADMPerturbativeSolve`"];
(* Exported symbols added here with SymbolName::usage *)  

ComplexFrequencyMatrix::usage="";

SteadyStatePerturbativeSolve::usage="";

TDPerturbativeSolve::usage="";

Begin["`Private`"]; (* Begin Private Context *)

(*SteadyStatePerturbativeSolve*)

ComplexFrequencyMatrix[H0_,\[CapitalGamma]_]:=Table[(H0[[i,i]]-H0[[j,j]])-I/2(\[CapitalGamma][[i,i]]+\[CapitalGamma][[j,j]]),{i,Length[H0]},{j,Length[H0]}];

DiagonalMatrixQ[expr_]:=MatrixQ[expr]&&expr===DiagonalMatrix[Tr[expr,List]];

SteadyStatePerturbativeSolve[(V_)?MatrixQ, (L_)?
   MatrixQ, (W_)?MatrixQ, FO_?(ArrayQ[#,4]&), {m_, n_}, {0}] :=
    -((I*L[[m, n]])/W[[m, n]]);

SteadyStatePerturbativeSolve[(V_)?MatrixQ, (L_)?
   MatrixQ, (W_)?MatrixQ, FO_?(ArrayQ[#,4]&), {m_, n_}, {1}] :=
    Sum[-((I*V[[p[0], n]]*L[[m, p[0]]])/(W[[m,n]]*W[[m, p[0]]])) + 
        (I*V[[m, p[0]]]*L[[p[0], n]])/(W[[m,n]]*W[[p[0], n]]), 
    	{p[0], 1, Length[V]}
    ];

SteadyStatePerturbativeSolve[(V_)?MatrixQ, (L_)?
   MatrixQ, (W_)?MatrixQ, 
     FO_?(ArrayQ[#,4]&), {m_, n_}, {2}] :=
    Module[ {FOelems = Most[ArrayRules[SparseArray[FO]][[All, 1]]]},
        Sum[If[ MemberQ[FOelems, {m, n, s[0], r[0]}],
                Sum[(V[[p[2], s[0]]]*
                      V[[r[0], p[1]]]*L[[p[1], 
                        p[2]]])/(W[[m, n]]*W[[p[1], p[2]]]*
                                 W[[p[1], s[0]]]*W[[r[0], 
                        s[0]]]) + (V[[p[1], s[0]]]*
                      V[[r[0], p[2]]]*L[[p[2], p[1]]])/
                              (W[[m, n]]*W[[p[2], 
                        p[1]]]*W[[r[0], p[1]]]*W[[r[0], s[0]]]), 
                          {p[2], 1, Length[V]}, {p[1], 1, Length[V]}]*
                 FO[[m, n, s[0], r[0]]],
                0
            ], {r[0], 1, Length[V]}, 
            {s[0], 1, Length[V]}] + 
        Sum[-((I*V[[p[0], n]]*V[[p[1], p[0]]]*L[[m, p[1]]])/
                   (W[[m, n]]*W[[m, p[0]]]*W[[m, 
               p[1]]])) + 
              (I*V[[m, p[0]]]*
           V[[p[1], n]]*L[[p[0], p[1]]])/(W[[m, 
             n]]*W[[p[0], n]]*
                   W[[p[0], p[1]]]) - (I*V[[m, p[0]]]*
           V[[p[0], p[1]]]*L[[p[1], n]])/
                (W[[m, n]]*W[[p[0], n]]*W[[p[1], 
             n]]) + (I*V[[m, p[1]]]*
           V[[p[0], n]]*L[[p[1], p[0]]])/
                (W[[m, n]]*W[[m, p[0]]]*W[[p[1], 
             p[0]]]), {p[1], 1, Length[V]}, {p[0], 1, Length[V]}]
    ];

SteadyStatePerturbativeSolve[(V_)?MatrixQ, (L_)?
   MatrixQ, (W_)?MatrixQ, 
     FO_?(ArrayQ[#,4]&), {m_, n_}, {3}] := 
   Module[{FOelems = Most[ArrayRules[SparseArray[FO]][[All, 1]]]}, 
     Sum[If[MemberQ[FOelems, {m, n, s[0], r[0]}], 
           
     Sum[(V[[p[2], s[0]]]*V[[p[3], p[2]]]*
           V[[r[0], p[1]]]*L[[p[1], p[3]]])/
                   (W[[m, n]]*W[[p[1], 
             p[2]]]*W[[p[1], p[3]]]*W[[p[1], 
             s[0]]]*W[[r[0], s[0]]]) - 
                 (V[[p[1], p[2]]]*V[[p[3], s[0]]]*
           V[[r[0], p[1]]]*L[[p[2], p[3]]])/
                   (W[[m, n]]*W[[p[1], 
             s[0]]]*W[[p[2], p[3]]]*W[[p[2], 
             s[0]]]*W[[r[0], s[0]]]) - 
                 (V[[p[1], p[3]]]*V[[p[2], s[0]]]*
           V[[r[0], p[1]]]*L[[p[3], p[2]]])/
                   (W[[m, n]]*W[[p[1], 
             p[2]]]*W[[p[1], s[0]]]*W[[p[3], 
             p[2]]]*W[[r[0], s[0]]]) + 
                 (V[[p[1], s[0]]]*V[[p[3], p[1]]]*
           V[[r[0], p[2]]]*L[[p[2], p[3]]])/
                   (W[[m, n]]*W[[p[2], 
             p[1]]]*W[[p[2], p[3]]]*W[[r[0], 
             p[1]]]*W[[r[0], s[0]]]) - 
                 (V[[p[1], s[0]]]*V[[p[2], p[3]]]*
           V[[r[0], p[2]]]*L[[p[3], p[1]]])/
                   (W[[m, n]]*W[[p[2], 
             p[1]]]*W[[p[3], p[1]]]*W[[r[0], 
             p[1]]]*W[[r[0], s[0]]]) + 
                 (V[[p[1], s[0]]]*V[[p[2], p[1]]]*
           V[[r[0], p[3]]]*L[[p[3], p[2]]])/
                   (W[[m, n]]*W[[p[3], 
             p[2]]]*W[[r[0], p[1]]]*W[[r[0], 
             p[2]]]*W[[r[0], s[0]]]), 
               {p[3], 1, Length[V]}, {p[2], 1, Length[V]}, {p[1], 1, 
        Length[V]}]*FO[[m, n, s[0], r[0]]], 0], 
         {r[0], 1, Length[V]}, {s[0], 1, Length[V]}] + 
       Sum[If[MemberQ[FOelems, {m, p[0], s[1], r[1]}], 
             
      Sum[(V[[p[0], n]]*V[[p[3], s[1]]]*
            V[[r[1], p[2]]]*L[[p[2], p[3]]])/
                     (W[[m, n]]*W[[m, 
              p[0]]]*W[[p[2], p[3]]]*W[[p[2], 
              s[1]]]*W[[r[1], s[1]]]) + 
                   (V[[p[0], n]]*V[[p[2], s[1]]]*
            V[[r[1], p[3]]]*L[[p[3], 
              p[2]]])/(W[[m, n]]*W[[m, p[0]]]*
                        W[[p[3], p[2]]]*W[[r[1], 
              p[2]]]*W[[r[1], s[1]]]), {p[3], 1, Length[V]}, 
                 {p[2], 1, Length[V]}]*FO[[m, p[0], s[1], r[1]]], 0] + 
           If[MemberQ[FOelems, {p[0], n, s[1], r[1]}], 
             
      Sum[-((V[[m, p[0]]]*V[[p[3], s[1]]]*
              V[[r[1], p[2]]]*L[[p[2], p[3]]])/
                        (W[[m, n]]*W[[p[0], 
                n]]*W[[p[2], p[3]]]*W[[p[2], 
                s[1]]]*W[[r[1], s[1]]])) - 
                   (V[[m, p[0]]]*V[[p[2], s[1]]]*
            V[[r[1], p[3]]]*L[[p[3], 
              p[2]]])/(W[[m, n]]*W[[p[0], n]]*
                        W[[p[3], p[2]]]*W[[r[1], 
              p[2]]]*W[[r[1], s[1]]]), {p[3], 1, Length[V]}, 
                 {p[2], 1, Length[V]}]*FO[[p[0], n, s[1], r[1]]], 
      0], {r[1], 1, Length[V]}, 
         {s[1], 1, Length[V]}, {p[0], 1, Length[V]}] + 
       Sum[-((I*V[[p[0], n]]*V[[p[1], p[0]]]*
          V[[p[2], p[1]]]*L[[m, p[2]]])/(W[[m, 
            n]]*W[[m, p[0]]]*
                   W[[m, p[1]]]*W[[m, p[2]]])) + (I*
        V[[m, p[0]]]*V[[p[1], n]]*V[[p[2], p[1]]]*
                L[[p[0], p[2]]])/(W[[m, 
          n]]*W[[p[0], n]]*W[[p[0], 
          p[1]]]*W[[p[0], p[2]]]) - 
           (I*V[[m, p[0]]]*V[[p[0], p[1]]]*
        V[[p[2], n]]*L[[p[1], p[2]]])/(W[[m, 
          n]]*W[[p[0], n]]*
                W[[p[1], n]]*W[[p[1], p[2]]]) + (I*
        V[[m, p[1]]]*V[[p[0], n]]*V[[p[2], p[0]]]*
                L[[p[1], p[2]]])/(W[[m, 
          n]]*W[[m, p[0]]]*W[[p[1], 
          p[0]]]*W[[p[1], p[2]]]) + 
           (I*V[[m, p[0]]]*V[[p[0], p[1]]]*
        V[[p[1], p[2]]]*L[[p[2], n]])/(W[[m, 
          n]]*W[[p[0], n]]*
                W[[p[1], n]]*W[[p[2], n]]) - (I*
        V[[m, p[1]]]*V[[p[0], n]]*V[[p[1], p[2]]]*
                L[[p[2], p[0]]])/(W[[m, 
          n]]*W[[m, p[0]]]*W[[p[1], 
          p[0]]]*W[[p[2], p[0]]]) + 
           (I*V[[m, p[2]]]*V[[p[0], n]]*
        V[[p[1], p[0]]]*L[[p[2], p[1]]])/(W[[m,
           n]]*W[[m, p[0]]]*
                W[[m, p[1]]]*W[[p[2], p[1]]]) - (I*
        V[[m, p[0]]]*V[[p[0], p[2]]]*V[[p[1], n]]*
                L[[p[2], p[1]]])/(W[[m, 
          n]]*W[[p[0], n]]*W[[p[0], 
          p[1]]]*W[[p[2], p[1]]]), 
         {p[2], 1, Length[V]}, {p[1], 1, Length[V]}, {p[0], 1, 
     Length[V]}]];

SteadyStatePerturbativeSolve[V_?MatrixQ, L_?MatrixQ, W_?MatrixQ, FO_?(ArrayQ[#,4]&), {m_, n_}, {4}] := 
  Module[{FOelems = Most[ArrayRules[SparseArray[FO]][[All,1]]]}, 
   Sum[(Sum[If[MemberQ[FOelems, {p[1], p[2], s[3], r[3]}], 
          Sum[(I*V[[p[2],s[0]]]*V[[p[5],s[3]]]*V[[r[0],p[1]]]*
              V[[r[3],p[4]]]*L[[p[4],p[5]]])/(W[[m,n]]*
              W[[p[1],p[2]]]*W[[p[1],s[0]]]*W[[p[4],p[5]]]*
              W[[p[4],s[3]]]*W[[r[0],s[0]]]*W[[r[3],s[3]]]), 
            {p[5], 1, Length[V]}, {p[4], 1, Length[V]}]*
           FO[[p[1],p[2],s[3],r[3]]], 0] + 
         If[MemberQ[FOelems, {p[1], p[2], s[3], r[3]}], 
          Sum[(I*V[[p[2],s[0]]]*V[[p[4],s[3]]]*V[[r[0],p[1]]]*
              V[[r[3],p[5]]]*L[[p[5],p[4]]])/(W[[m,n]]*
              W[[p[1],p[2]]]*W[[p[1],s[0]]]*W[[p[5],p[4]]]*
              W[[r[0],s[0]]]*W[[r[3],p[4]]]*W[[r[3],s[3]]]), 
            {p[5], 1, Length[V]}, {p[4], 1, Length[V]}]*
           FO[[p[1],p[2],s[3],r[3]]], 0] + 
         If[MemberQ[FOelems, {p[2], p[1], s[3], r[3]}], 
          Sum[(I*V[[p[1],s[0]]]*V[[p[5],s[3]]]*V[[r[0],p[2]]]*
              V[[r[3],p[4]]]*L[[p[4],p[5]]])/(W[[m,n]]*
              W[[p[2],p[1]]]*W[[p[4],p[5]]]*W[[p[4],s[3]]]*
              W[[r[0],p[1]]]*W[[r[0],s[0]]]*W[[r[3],s[3]]]), 
            {p[5], 1, Length[V]}, {p[4], 1, Length[V]}]*
           FO[[p[2],p[1],s[3],r[3]]], 0] + 
         If[MemberQ[FOelems, {p[2], p[1], s[3], r[3]}], 
          Sum[(I*V[[p[1],s[0]]]*V[[p[4],s[3]]]*V[[r[0],p[2]]]*
              V[[r[3],p[5]]]*L[[p[5],p[4]]])/(W[[m,n]]*
              W[[p[2],p[1]]]*W[[p[5],p[4]]]*W[[r[0],p[1]]]*
              W[[r[0],s[0]]]*W[[r[3],p[4]]]*W[[r[3],s[3]]]), 
            {p[5], 1, Length[V]}, {p[4], 1, Length[V]}]*
           FO[[p[2],p[1],s[3],r[3]]], 0], {r[3], 1, Length[V]}, 
        {s[3], 1, Length[V]}, {p[2], 1, Length[V]}, 
        {p[1], 1, Length[V]}] + 
       Sum[(V[[p[2],s[0]]]*V[[p[3],p[2]]]*V[[p[4],p[3]]]*
           V[[r[0],p[1]]]*L[[p[1],p[4]]])/(W[[m,n]]*
           W[[p[1],p[2]]]*W[[p[1],p[3]]]*W[[p[1],p[4]]]*
           W[[p[1],s[0]]]*W[[r[0],s[0]]]) - 
         (V[[p[1],p[2]]]*V[[p[3],s[0]]]*V[[p[4],p[3]]]*V[[r[0],p[1]]]*
           L[[p[2],p[4]]])/(W[[m,n]]*W[[p[1],s[0]]]*
           W[[p[2],p[3]]]*W[[p[2],p[4]]]*W[[p[2],s[0]]]*
           W[[r[0],s[0]]]) - (V[[p[1],p[3]]]*V[[p[2],s[0]]]*
           V[[p[4],p[2]]]*V[[r[0],p[1]]]*L[[p[3],p[4]]])/
          (W[[m,n]]*W[[p[1],p[2]]]*W[[p[1],s[0]]]*
           W[[p[3],p[2]]]*W[[p[3],p[4]]]*W[[r[0],s[0]]]) + 
         (V[[p[1],p[2]]]*V[[p[2],p[3]]]*V[[p[4],s[0]]]*V[[r[0],p[1]]]*
           L[[p[3],p[4]]])/(W[[m,n]]*W[[p[1],s[0]]]*
           W[[p[2],s[0]]]*W[[p[3],p[4]]]*W[[p[3],s[0]]]*
           W[[r[0],s[0]]]) + (V[[p[1],p[3]]]*V[[p[2],s[0]]]*
           V[[p[3],p[4]]]*V[[r[0],p[1]]]*L[[p[4],p[2]]])/
          (W[[m,n]]*W[[p[1],p[2]]]*W[[p[1],s[0]]]*
           W[[p[3],p[2]]]*W[[p[4],p[2]]]*W[[r[0],s[0]]]) - 
         (V[[p[1],p[4]]]*V[[p[2],s[0]]]*V[[p[3],p[2]]]*V[[r[0],p[1]]]*
           L[[p[4],p[3]]])/(W[[m,n]]*W[[p[1],p[2]]]*
           W[[p[1],p[3]]]*W[[p[1],s[0]]]*W[[p[4],p[3]]]*
           W[[r[0],s[0]]]) + (V[[p[1],p[2]]]*V[[p[2],p[4]]]*
           V[[p[3],s[0]]]*V[[r[0],p[1]]]*L[[p[4],p[3]]])/
          (W[[m,n]]*W[[p[1],s[0]]]*W[[p[2],p[3]]]*
           W[[p[2],s[0]]]*W[[p[4],p[3]]]*W[[r[0],s[0]]]) + 
         (V[[p[1],s[0]]]*V[[p[3],p[1]]]*V[[p[4],p[3]]]*V[[r[0],p[2]]]*
           L[[p[2],p[4]]])/(W[[m,n]]*W[[p[2],p[1]]]*
           W[[p[2],p[3]]]*W[[p[2],p[4]]]*W[[r[0],p[1]]]*
           W[[r[0],s[0]]]) - (V[[p[1],s[0]]]*V[[p[2],p[3]]]*
           V[[p[4],p[1]]]*V[[r[0],p[2]]]*L[[p[3],p[4]]])/
          (W[[m,n]]*W[[p[2],p[1]]]*W[[p[3],p[1]]]*
           W[[p[3],p[4]]]*W[[r[0],p[1]]]*W[[r[0],s[0]]]) + 
         (V[[p[1],s[0]]]*V[[p[2],p[3]]]*V[[p[3],p[4]]]*V[[r[0],p[2]]]*
           L[[p[4],p[1]]])/(W[[m,n]]*W[[p[2],p[1]]]*
           W[[p[3],p[1]]]*W[[p[4],p[1]]]*W[[r[0],p[1]]]*
           W[[r[0],s[0]]]) - (V[[p[1],s[0]]]*V[[p[2],p[4]]]*
           V[[p[3],p[1]]]*V[[r[0],p[2]]]*L[[p[4],p[3]]])/
          (W[[m,n]]*W[[p[2],p[1]]]*W[[p[2],p[3]]]*
           W[[p[4],p[3]]]*W[[r[0],p[1]]]*W[[r[0],s[0]]]) + 
         (V[[p[1],s[0]]]*V[[p[2],p[1]]]*V[[p[4],p[2]]]*V[[r[0],p[3]]]*
           L[[p[3],p[4]]])/(W[[m,n]]*W[[p[3],p[2]]]*
           W[[p[3],p[4]]]*W[[r[0],p[1]]]*W[[r[0],p[2]]]*
           W[[r[0],s[0]]]) - (V[[p[1],s[0]]]*V[[p[2],p[1]]]*
           V[[p[3],p[4]]]*V[[r[0],p[3]]]*L[[p[4],p[2]]])/
          (W[[m,n]]*W[[p[3],p[2]]]*W[[p[4],p[2]]]*
           W[[r[0],p[1]]]*W[[r[0],p[2]]]*W[[r[0],s[0]]]) + 
         (V[[p[1],s[0]]]*V[[p[2],p[1]]]*V[[p[3],p[2]]]*V[[r[0],p[4]]]*
           L[[p[4],p[3]]])/(W[[m,n]]*W[[p[4],p[3]]]*
           W[[r[0],p[1]]]*W[[r[0],p[2]]]*W[[r[0],p[3]]]*
           W[[r[0],s[0]]]), {p[4], 1, Length[V]}, 
        {p[3], 1, Length[V]}, {p[2], 1, Length[V]}, 
        {p[1], 1, Length[V]}])*FO[[m,n,s[0],r[0]]], 
     {r[0], 1, Length[V]}, {s[0], 1, Length[V]}] + 
    Sum[If[MemberQ[FOelems, {m, p[0], s[1], r[1]}], 
       Sum[(V[[p[0],n]]*V[[p[3],s[1]]]*V[[p[4],p[3]]]*V[[r[1],p[2]]]*
            L[[p[2],p[4]]])/(W[[m,n]]*W[[m,p[0]]]*
            W[[p[2],p[3]]]*W[[p[2],p[4]]]*W[[p[2],s[1]]]*
            W[[r[1],s[1]]]) - (V[[p[0],n]]*V[[p[2],p[3]]]*
            V[[p[4],s[1]]]*V[[r[1],p[2]]]*L[[p[3],p[4]]])/
           (W[[m,n]]*W[[m,p[0]]]*W[[p[2],s[1]]]*
            W[[p[3],p[4]]]*W[[p[3],s[1]]]*W[[r[1],s[1]]]) - 
          (V[[p[0],n]]*V[[p[2],p[4]]]*V[[p[3],s[1]]]*V[[r[1],p[2]]]*
            L[[p[4],p[3]]])/(W[[m,n]]*W[[m,p[0]]]*
            W[[p[2],p[3]]]*W[[p[2],s[1]]]*W[[p[4],p[3]]]*
            W[[r[1],s[1]]]) + (V[[p[0],n]]*V[[p[2],s[1]]]*
            V[[p[4],p[2]]]*V[[r[1],p[3]]]*L[[p[3],p[4]]])/
           (W[[m,n]]*W[[m,p[0]]]*W[[p[3],p[2]]]*
            W[[p[3],p[4]]]*W[[r[1],p[2]]]*W[[r[1],s[1]]]) - 
          (V[[p[0],n]]*V[[p[2],s[1]]]*V[[p[3],p[4]]]*V[[r[1],p[3]]]*
            L[[p[4],p[2]]])/(W[[m,n]]*W[[m,p[0]]]*
            W[[p[3],p[2]]]*W[[p[4],p[2]]]*W[[r[1],p[2]]]*
            W[[r[1],s[1]]]) + (V[[p[0],n]]*V[[p[2],s[1]]]*
            V[[p[3],p[2]]]*V[[r[1],p[4]]]*L[[p[4],p[3]]])/
           (W[[m,n]]*W[[m,p[0]]]*W[[p[4],p[3]]]*
            W[[r[1],p[2]]]*W[[r[1],p[3]]]*W[[r[1],s[1]]]), 
         {p[4], 1, Length[V]}, {p[3], 1, Length[V]}, 
         {p[2], 1, Length[V]}]*FO[[m,p[0],s[1],r[1]]], 0] + 
      If[MemberQ[FOelems, {p[0], n, s[1], r[1]}], 
       Sum[-((V[[m,p[0]]]*V[[p[3],s[1]]]*V[[p[4],p[3]]]*V[[r[1],p[2]]]*
             L[[p[2],p[4]]])/(W[[m,n]]*W[[p[0],n]]*
             W[[p[2],p[3]]]*W[[p[2],p[4]]]*W[[p[2],s[1]]]*
             W[[r[1],s[1]]])) + (V[[m,p[0]]]*V[[p[2],p[3]]]*
            V[[p[4],s[1]]]*V[[r[1],p[2]]]*L[[p[3],p[4]]])/
           (W[[m,n]]*W[[p[0],n]]*W[[p[2],s[1]]]*
            W[[p[3],p[4]]]*W[[p[3],s[1]]]*W[[r[1],s[1]]]) + 
          (V[[m,p[0]]]*V[[p[2],p[4]]]*V[[p[3],s[1]]]*V[[r[1],p[2]]]*
            L[[p[4],p[3]]])/(W[[m,n]]*W[[p[0],n]]*
            W[[p[2],p[3]]]*W[[p[2],s[1]]]*W[[p[4],p[3]]]*
            W[[r[1],s[1]]]) - (V[[m,p[0]]]*V[[p[2],s[1]]]*
            V[[p[4],p[2]]]*V[[r[1],p[3]]]*L[[p[3],p[4]]])/
           (W[[m,n]]*W[[p[0],n]]*W[[p[3],p[2]]]*
            W[[p[3],p[4]]]*W[[r[1],p[2]]]*W[[r[1],s[1]]]) + 
          (V[[m,p[0]]]*V[[p[2],s[1]]]*V[[p[3],p[4]]]*V[[r[1],p[3]]]*
            L[[p[4],p[2]]])/(W[[m,n]]*W[[p[0],n]]*
            W[[p[3],p[2]]]*W[[p[4],p[2]]]*W[[r[1],p[2]]]*
            W[[r[1],s[1]]]) - (V[[m,p[0]]]*V[[p[2],s[1]]]*
            V[[p[3],p[2]]]*V[[r[1],p[4]]]*L[[p[4],p[3]]])/
           (W[[m,n]]*W[[p[0],n]]*W[[p[4],p[3]]]*
            W[[r[1],p[2]]]*W[[r[1],p[3]]]*W[[r[1],s[1]]]), 
         {p[4], 1, Length[V]}, {p[3], 1, Length[V]}, 
         {p[2], 1, Length[V]}]*FO[[p[0],n,s[1],r[1]]], 0], 
     {r[1], 1, Length[V]}, {s[1], 1, Length[V]}, 
     {p[0], 1, Length[V]}] + 
    Sum[If[MemberQ[FOelems, {m, p[1], s[2], r[2]}], 
       Sum[(V[[p[0],n]]*V[[p[1],p[0]]]*V[[p[4],s[2]]]*V[[r[2],p[3]]]*
            L[[p[3],p[4]]])/(W[[m,n]]*W[[m,p[0]]]*
            W[[m,p[1]]]*W[[p[3],p[4]]]*W[[p[3],s[2]]]*
            W[[r[2],s[2]]]) + (V[[p[0],n]]*V[[p[1],p[0]]]*
            V[[p[3],s[2]]]*V[[r[2],p[4]]]*L[[p[4],p[3]]])/
           (W[[m,n]]*W[[m,p[0]]]*W[[m,p[1]]]*
            W[[p[4],p[3]]]*W[[r[2],p[3]]]*W[[r[2],s[2]]]), 
         {p[4], 1, Length[V]}, {p[3], 1, Length[V]}]*
        FO[[m,p[1],s[2],r[2]]], 0] + 
      If[MemberQ[FOelems, {p[0], p[1], s[2], r[2]}], 
       Sum[-((V[[m,p[0]]]*V[[p[1],n]]*V[[p[4],s[2]]]*V[[r[2],p[3]]]*
             L[[p[3],p[4]]])/(W[[m,n]]*W[[p[0],n]]*
             W[[p[0],p[1]]]*W[[p[3],p[4]]]*W[[p[3],s[2]]]*
             W[[r[2],s[2]]])) - (V[[m,p[0]]]*V[[p[1],n]]*
            V[[p[3],s[2]]]*V[[r[2],p[4]]]*L[[p[4],p[3]]])/
           (W[[m,n]]*W[[p[0],n]]*W[[p[0],p[1]]]*
            W[[p[4],p[3]]]*W[[r[2],p[3]]]*W[[r[2],s[2]]]), 
         {p[4], 1, Length[V]}, {p[3], 1, Length[V]}]*
        FO[[p[0],p[1],s[2],r[2]]], 0] + 
      If[MemberQ[FOelems, {p[1], n, s[2], r[2]}], 
       Sum[(V[[m,p[0]]]*V[[p[0],p[1]]]*V[[p[4],s[2]]]*V[[r[2],p[3]]]*
            L[[p[3],p[4]]])/(W[[m,n]]*W[[p[0],n]]*
            W[[p[1],n]]*W[[p[3],p[4]]]*W[[p[3],s[2]]]*
            W[[r[2],s[2]]]) + (V[[m,p[0]]]*V[[p[0],p[1]]]*
            V[[p[3],s[2]]]*V[[r[2],p[4]]]*L[[p[4],p[3]]])/
           (W[[m,n]]*W[[p[0],n]]*W[[p[1],n]]*
            W[[p[4],p[3]]]*W[[r[2],p[3]]]*W[[r[2],s[2]]]), 
         {p[4], 1, Length[V]}, {p[3], 1, Length[V]}]*
        FO[[p[1],n,s[2],r[2]]], 0] + 
      If[MemberQ[FOelems, {p[1], p[0], s[2], r[2]}], 
       Sum[-((V[[m,p[1]]]*V[[p[0],n]]*V[[p[4],s[2]]]*V[[r[2],p[3]]]*
             L[[p[3],p[4]]])/(W[[m,n]]*W[[m,p[0]]]*
             W[[p[1],p[0]]]*W[[p[3],p[4]]]*W[[p[3],s[2]]]*
             W[[r[2],s[2]]])) - (V[[m,p[1]]]*V[[p[0],n]]*
            V[[p[3],s[2]]]*V[[r[2],p[4]]]*L[[p[4],p[3]]])/
           (W[[m,n]]*W[[m,p[0]]]*W[[p[1],p[0]]]*
            W[[p[4],p[3]]]*W[[r[2],p[3]]]*W[[r[2],s[2]]]), 
         {p[4], 1, Length[V]}, {p[3], 1, Length[V]}]*
        FO[[p[1],p[0],s[2],r[2]]], 0], {r[2], 1, Length[V]}, 
     {s[2], 1, Length[V]}, {p[1], 1, Length[V]}, 
     {p[0], 1, Length[V]}] + 
    Sum[-((I*V[[p[0],n]]*V[[p[1],p[0]]]*V[[p[2],p[1]]]*V[[p[3],p[2]]]*
         L[[m,p[3]]])/(W[[m,n]]*W[[m,p[0]]]*W[[m,p[1]]]*
         W[[m,p[2]]]*W[[m,p[3]]])) + 
      (I*V[[m,p[0]]]*V[[p[1],n]]*V[[p[2],p[1]]]*V[[p[3],p[2]]]*
        L[[p[0],p[3]]])/(W[[m,n]]*W[[p[0],n]]*
        W[[p[0],p[1]]]*W[[p[0],p[2]]]*W[[p[0],p[3]]]) - 
      (I*V[[m,p[0]]]*V[[p[0],p[1]]]*V[[p[2],n]]*V[[p[3],p[2]]]*
        L[[p[1],p[3]]])/(W[[m,n]]*W[[p[0],n]]*W[[p[1],n]]*
        W[[p[1],p[2]]]*W[[p[1],p[3]]]) + 
      (I*V[[m,p[1]]]*V[[p[0],n]]*V[[p[2],p[0]]]*V[[p[3],p[2]]]*
        L[[p[1],p[3]]])/(W[[m,n]]*W[[m,p[0]]]*
        W[[p[1],p[0]]]*W[[p[1],p[2]]]*W[[p[1],p[3]]]) + 
      (I*V[[m,p[0]]]*V[[p[0],p[1]]]*V[[p[1],p[2]]]*V[[p[3],n]]*
        L[[p[2],p[3]]])/(W[[m,n]]*W[[p[0],n]]*W[[p[1],n]]*
        W[[p[2],n]]*W[[p[2],p[3]]]) - 
      (I*V[[m,p[1]]]*V[[p[0],n]]*V[[p[1],p[2]]]*V[[p[3],p[0]]]*
        L[[p[2],p[3]]])/(W[[m,n]]*W[[m,p[0]]]*
        W[[p[1],p[0]]]*W[[p[2],p[0]]]*W[[p[2],p[3]]]) + 
      (I*V[[m,p[2]]]*V[[p[0],n]]*V[[p[1],p[0]]]*V[[p[3],p[1]]]*
        L[[p[2],p[3]]])/(W[[m,n]]*W[[m,p[0]]]*W[[m,p[1]]]*
        W[[p[2],p[1]]]*W[[p[2],p[3]]]) - 
      (I*V[[m,p[0]]]*V[[p[0],p[2]]]*V[[p[1],n]]*V[[p[3],p[1]]]*
        L[[p[2],p[3]]])/(W[[m,n]]*W[[p[0],n]]*
        W[[p[0],p[1]]]*W[[p[2],p[1]]]*W[[p[2],p[3]]]) - 
      (I*V[[m,p[0]]]*V[[p[0],p[1]]]*V[[p[1],p[2]]]*V[[p[2],p[3]]]*
        L[[p[3],n]])/(W[[m,n]]*W[[p[0],n]]*W[[p[1],n]]*
        W[[p[2],n]]*W[[p[3],n]]) + 
      (I*V[[m,p[1]]]*V[[p[0],n]]*V[[p[1],p[2]]]*V[[p[2],p[3]]]*
        L[[p[3],p[0]]])/(W[[m,n]]*W[[m,p[0]]]*
        W[[p[1],p[0]]]*W[[p[2],p[0]]]*W[[p[3],p[0]]]) - 
      (I*V[[m,p[2]]]*V[[p[0],n]]*V[[p[1],p[0]]]*V[[p[2],p[3]]]*
        L[[p[3],p[1]]])/(W[[m,n]]*W[[m,p[0]]]*W[[m,p[1]]]*
        W[[p[2],p[1]]]*W[[p[3],p[1]]]) + 
      (I*V[[m,p[0]]]*V[[p[0],p[2]]]*V[[p[1],n]]*V[[p[2],p[3]]]*
        L[[p[3],p[1]]])/(W[[m,n]]*W[[p[0],n]]*
        W[[p[0],p[1]]]*W[[p[2],p[1]]]*W[[p[3],p[1]]]) + 
      (I*V[[m,p[3]]]*V[[p[0],n]]*V[[p[1],p[0]]]*V[[p[2],p[1]]]*
        L[[p[3],p[2]]])/(W[[m,n]]*W[[m,p[0]]]*W[[m,p[1]]]*
        W[[m,p[2]]]*W[[p[3],p[2]]]) - 
      (I*V[[m,p[0]]]*V[[p[0],p[3]]]*V[[p[1],n]]*V[[p[2],p[1]]]*
        L[[p[3],p[2]]])/(W[[m,n]]*W[[p[0],n]]*
        W[[p[0],p[1]]]*W[[p[0],p[2]]]*W[[p[3],p[2]]]) + 
      (I*V[[m,p[0]]]*V[[p[0],p[1]]]*V[[p[1],p[3]]]*V[[p[2],n]]*
        L[[p[3],p[2]]])/(W[[m,n]]*W[[p[0],n]]*W[[p[1],n]]*
        W[[p[1],p[2]]]*W[[p[3],p[2]]]) - 
      (I*V[[m,p[1]]]*V[[p[0],n]]*V[[p[1],p[3]]]*V[[p[2],p[0]]]*
        L[[p[3],p[2]]])/(W[[m,n]]*W[[m,p[0]]]*
        W[[p[1],p[0]]]*W[[p[1],p[2]]]*W[[p[3],p[2]]]), 
     {p[3], 1, Length[V]}, {p[2], 1, Length[V]}, {p[1], 1, Length[V]}, 
     {p[0], 1, Length[V]}]];

SteadyStatePerturbativeSolve[V_?MatrixQ,L_?MatrixQ,W_?MatrixQ,FO_?(ArrayQ[#,4]&),{i_}]:=
	Table[SteadyStatePerturbativeSolve[V,L,W,FO,{m,n},{i}],{m,Length[V]},{n,Length[V]}];

SteadyStatePerturbativeSolve[V_?MatrixQ,L_?MatrixQ,W_?MatrixQ,FO_?(ArrayQ[#,4]&),i_]:=
	Sum[SteadyStatePerturbativeSolve[V,L,W,FO,{j}],{j,0,i}];

SteadyStatePerturbativeSolve[V_?MatrixQ, L_?MatrixQ, W_?MatrixQ,  {m_?IntegerQ, n_?IntegerQ}, {i_?IntegerQ}] :=
	SteadyStatePerturbativeSolve[V, L, W, ConstantArray[0,ConstantArray[Dimensions[V][[1]],4]], {m, n}, {i}];

SteadyStatePerturbativeSolve[V_?MatrixQ,L_?MatrixQ,W_?MatrixQ,c:i_?IntegerQ|{i_?IntegerQ}]:=
	SteadyStatePerturbativeSolve[V,L,W,ConstantArray[0,ConstantArray[Dimensions[V][[1]],4]],c];

(*TDPerturbativeSolve*)

TDPerturbativeSolve[V_,\[Rho]0_,\[Omega]_,FO_,t1_,{m_,n_},{0}]:=\[Rho]0[[m,n]]/E^(I*t1*\[Omega][[m,n]]);

TDPerturbativeSolve[V_,\[Rho]0_,\[Omega]_,FO_,t1_,{m_,n_},{1}]:=Sum[((V[[m,p]]*\[Rho]0[[p,n]])/\[Omega][[m,p]]-(E^(I*t1*\[Omega][[m,p]])*V[[m,p]]*\[Rho]0[[p,n]])/\[Omega][[m,p]]-(V[[p,n]]*\[Rho]0[[m,p]])/\[Omega][[p,n]]+(E^(I*t1*\[Omega][[p,n]])*V[[p,n]]*\[Rho]0[[m,p]])/\[Omega][[p,n]])/E^(I*t1*\[Omega][[m,n]]),{p,Length[V]}];

TDPerturbativeSolve[V_,\[Rho]0_,\[Omega]_,FO_,t1_,{m_,n_},{2}]:=Sum[If[m===r&&n===s,0,(((-I)*FO[[m,n,s,r]]*V[[p2,s]]*V[[r,p1]]*\[Rho]0[[p1,p2]])/((\[Omega][[m,n]]+\[Omega][[p2,p1]])*\[Omega][[p2,s]]*(\[Omega][[p2,s]]+\[Omega][[r,p1]]))-(I*E^(I*(t1*\[Omega][[m,n]]+t1*\[Omega][[p1,r]]))*FO[[m,n,s,r]]*V[[p1,s]]*V[[r,p2]]*\[Rho]0[[p2,p1]])/((\[Omega][[m,n]]+\[Omega][[p1,r]])*\[Omega][[p1,s]]*\[Omega][[r,p2]])-(I*FO[[m,n,s,r]]*V[[p1,s]]*V[[r,p2]]*\[Rho]0[[p2,p1]])/((\[Omega][[m,n]]+\[Omega][[p1,p2]])*\[Omega][[r,p2]]*(\[Omega][[p1,s]]+\[Omega][[r,p2]]))+I*((FO[[m,n,s,r]]*V[[p1,s]]*V[[r,p2]]*\[Rho]0[[p2,p1]])/((\[Omega][[m,n]]+\[Omega][[p1,r]])*\[Omega][[p1,s]]*\[Omega][[r,p2]])+(E^(I*(t1*\[Omega][[m,n]]+t1*\[Omega][[p1,p2]]))*FO[[m,n,s,r]]*V[[p1,s]]*V[[r,p2]]*\[Rho]0[[p2,p1]])/((\[Omega][[m,n]]+\[Omega][[p1,p2]])*\[Omega][[r,p2]]*(\[Omega][[p1,s]]+\[Omega][[r,p2]]))+(E^(I*t1*\[Omega][[m,n]]-I*t1*\[Omega][[r,s]])*FO[[m,n,s,r]]*V[[p1,s]]*V[[r,p2]]*\[Rho]0[[p2,p1]])/(\[Omega][[p1,s]]*\[Omega][[r,p2]]*(\[Omega][[m,n]]-\[Omega][[r,s]]))+(FO[[m,n,s,r]]*V[[p1,s]]*V[[r,p2]]*\[Rho]0[[p2,p1]])/(\[Omega][[r,p2]]*(\[Omega][[p1,s]]+\[Omega][[r,p2]])*(\[Omega][[m,n]]-\[Omega][[r,s]])))-(I*FO[[m,n,s,r]]*V[[p2,s]]*V[[r,p1]]*\[Rho]0[[p1,p2]])/(\[Omega][[p2,s]]*\[Omega][[r,p1]]*(\[Omega][[m,n]]-\[Omega][[r,s]]))-(I*E^(I*t1*\[Omega][[m,n]]-I*t1*\[Omega][[r,s]])*FO[[m,n,s,r]]*V[[p2,s]]*V[[r,p1]]*\[Rho]0[[p1,p2]])/(\[Omega][[p2,s]]*(\[Omega][[p2,s]]+\[Omega][[r,p1]])*(\[Omega][[m,n]]-\[Omega][[r,s]]))-(I*FO[[m,n,s,r]]*V[[p1,s]]*V[[r,p2]]*\[Rho]0[[p2,p1]])/(\[Omega][[p1,s]]*\[Omega][[r,p2]]*(\[Omega][[m,n]]-\[Omega][[r,s]]))-(I*E^(I*t1*\[Omega][[m,n]]-I*t1*\[Omega][[r,s]])*FO[[m,n,s,r]]*V[[p1,s]]*V[[r,p2]]*\[Rho]0[[p2,p1]])/(\[Omega][[r,p2]]*(\[Omega][[p1,s]]+\[Omega][[r,p2]])*(\[Omega][[m,n]]-\[Omega][[r,s]]))-(I*E^(I*(t1*\[Omega][[m,n]]+t1*\[Omega][[s,p1]]))*FO[[m,n,s,r]]*V[[p2,s]]*V[[r,p1]]*\[Rho]0[[p1,p2]])/(\[Omega][[p2,s]]*\[Omega][[r,p1]]*(\[Omega][[m,n]]+\[Omega][[s,p1]]))+I*((E^(I*(t1*\[Omega][[m,n]]+t1*\[Omega][[p2,p1]]))*FO[[m,n,s,r]]*V[[p2,s]]*V[[r,p1]]*\[Rho]0[[p1,p2]])/((\[Omega][[m,n]]+\[Omega][[p2,p1]])*\[Omega][[p2,s]]*(\[Omega][[p2,s]]+\[Omega][[r,p1]]))+(E^(I*t1*\[Omega][[m,n]]-I*t1*\[Omega][[r,s]])*FO[[m,n,s,r]]*V[[p2,s]]*V[[r,p1]]*\[Rho]0[[p1,p2]])/(\[Omega][[p2,s]]*\[Omega][[r,p1]]*(\[Omega][[m,n]]-\[Omega][[r,s]]))+(FO[[m,n,s,r]]*V[[p2,s]]*V[[r,p1]]*\[Rho]0[[p1,p2]])/(\[Omega][[p2,s]]*(\[Omega][[p2,s]]+\[Omega][[r,p1]])*(\[Omega][[m,n]]-\[Omega][[r,s]]))+(FO[[m,n,s,r]]*V[[p2,s]]*V[[r,p1]]*\[Rho]0[[p1,p2]])/(\[Omega][[p2,s]]*\[Omega][[r,p1]]*(\[Omega][[m,n]]+\[Omega][[s,p1]]))))/E^(I*t1*\[Omega][[m,n]])],{p2,Length[V]},{p1,Length[V]},{r,Length[V]},{s,Length[V]}]+Sum[((E^(I*t1*\[Omega][[p,n]])*V[[m,p1]]*V[[p,n]]*\[Rho]0[[p1,p]])/(\[Omega][[m,p1]]*(\[Omega][[m,p1]]+\[Omega][[p,n]]))-(E^(I*(t1*\[Omega][[m,p1]]+t1*\[Omega][[p,n]]))*V[[m,p1]]*V[[p,n]]*\[Rho]0[[p1,p]])/(\[Omega][[m,p1]]*(\[Omega][[m,p1]]+\[Omega][[p,n]]))-(V[[m,p1]]*V[[p,n]]*\[Rho]0[[p1,p]])/(\[Omega][[p,n]]*(\[Omega][[m,p1]]+\[Omega][[p,n]]))+(E^(I*t1*\[Omega][[p,n]])*V[[m,p1]]*V[[p,n]]*\[Rho]0[[p1,p]])/(\[Omega][[p,n]]*(\[Omega][[m,p1]]+\[Omega][[p,n]]))+(V[[m,p]]*V[[p,p1]]*\[Rho]0[[p1,n]])/(\[Omega][[m,p]]*(\[Omega][[m,p]]+\[Omega][[p,p1]]))-(E^(I*t1*\[Omega][[m,p]])*V[[m,p]]*V[[p,p1]]*\[Rho]0[[p1,n]])/(\[Omega][[m,p]]*(\[Omega][[m,p]]+\[Omega][[p,p1]]))-(E^(I*t1*\[Omega][[m,p]])*V[[m,p]]*V[[p,p1]]*\[Rho]0[[p1,n]])/(\[Omega][[p,p1]]*(\[Omega][[m,p]]+\[Omega][[p,p1]]))+(E^(I*(t1*\[Omega][[m,p]]+t1*\[Omega][[p,p1]]))*V[[m,p]]*V[[p,p1]]*\[Rho]0[[p1,n]])/(\[Omega][[p,p1]]*(\[Omega][[m,p]]+\[Omega][[p,p1]]))-(V[[m,p]]*V[[p1,n]]*\[Rho]0[[p,p1]])/(\[Omega][[m,p]]*(\[Omega][[m,p]]+\[Omega][[p1,n]]))+(E^(I*t1*\[Omega][[m,p]])*V[[m,p]]*V[[p1,n]]*\[Rho]0[[p,p1]])/(\[Omega][[m,p]]*(\[Omega][[m,p]]+\[Omega][[p1,n]]))+(E^(I*t1*\[Omega][[m,p]])*V[[m,p]]*V[[p1,n]]*\[Rho]0[[p,p1]])/(\[Omega][[p1,n]]*(\[Omega][[m,p]]+\[Omega][[p1,n]]))-(E^(I*(t1*\[Omega][[m,p]]+t1*\[Omega][[p1,n]]))*V[[m,p]]*V[[p1,n]]*\[Rho]0[[p,p1]])/(\[Omega][[p1,n]]*(\[Omega][[m,p]]+\[Omega][[p1,n]]))+(V[[p,n]]*V[[p1,p]]*\[Rho]0[[m,p1]])/(\[Omega][[p,n]]*(\[Omega][[p,n]]+\[Omega][[p1,p]]))-(E^(I*t1*\[Omega][[p,n]])*V[[p,n]]*V[[p1,p]]*\[Rho]0[[m,p1]])/(\[Omega][[p,n]]*(\[Omega][[p,n]]+\[Omega][[p1,p]]))-(E^(I*t1*\[Omega][[p,n]])*V[[p,n]]*V[[p1,p]]*\[Rho]0[[m,p1]])/(\[Omega][[p1,p]]*(\[Omega][[p,n]]+\[Omega][[p1,p]]))+(E^(I*(t1*\[Omega][[p,n]]+t1*\[Omega][[p1,p]]))*V[[p,n]]*V[[p1,p]]*\[Rho]0[[m,p1]])/(\[Omega][[p1,p]]*(\[Omega][[p,n]]+\[Omega][[p1,p]])))/E^(I*t1*\[Omega][[m,n]]),{p1,Length[V]},{p,Length[V]}];

TDPerturbativeSolve[V_,\[Rho]0_,\[Omega]_,FO_,t1_,{i_}]:=Table[TDPerturbativeSolve[V,\[Rho]0,\[Omega],FO,t1,{m,n},{i}],{m,Length[V]},{n,Length[V]}];

TDPerturbativeSolve[V_,\[Rho]0_,\[Omega]_,FO_,t1_,i_]:=Sum[TDPerturbativeSolve[V,\[Rho]0,\[Omega],FO,t1,{j}],{j,0,i}];

End[]; (* End Private Context *)

EndPackage[];
