(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2014 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`CouplingCoefficients`"];

CG::usage = "";

TJS::usage = "";

SJS::usage = "";

NJS::usage = "";

Begin["`Private`"];

(*Clebsch-Gordan coefficients*)

(*Triangular condition [Var88, Sec. 8.1.1, Eq. (1)]*)

CG[{F1_, m1_}, {F2_, m2_}, {F3_, m3_}] /; Abs[F1-F2] > F3||F1+F2 < F3 := CG[{F1, m1}, {F2, m2}, {F3, m3}] = 0;

(*Conservation of z-projection of angular momentum [Var88, Sec. 8.1.1, Eq. (2)]*)

CG[{F1_, m1_}, {F2_, m2_}, {F3_, m3_}] /; m1+m2 != m3 := CG[{F1, m1}, {F2, m2}, {F3, m3}] = 0;

(*z-projection is less than total angular momentum [Var88, Sec. 8.1.1, condition (c)]*)

CG[{F1_, m1_}, {F2_, m2_}, {F3_, m3_}] /; Abs[m1] > F1||Abs[m2] > F2||Abs[m3] > F3 := CG[{F1, m1}, {F2, m2}, {F3, m3}] = 0;

(*CG returns the value of the Clebsch-Gordan coefficient and stores the value for future use.*)

CG[{F1_, m1_}, {F2_, m2_}, {F3_, m3_}] := CG[{F1, m1}, {F2, m2}, {F3, m3}] = ClebschGordan[{F1, m1}, {F2, m2}, {F3, m3}];

(*Three-J Symbol*)

(*The conditions for the Three-J Symbol come from those for the Clebsch-Gordan coefficents.*)

(*Triangular condition*)

TJS[{F1_, m1_}, {F2_, m2_}, {F3_, m3_}] /; Abs[F1-F2] > F3||F1+F2 < F3 := TJS[{F1, m1}, {F2, m2}, {F3, m3}] = 0;

(*Conservation of z-projection of angular momentum*)

TJS[{F1_, m1_}, {F2_, m2_}, {F3_, m3_}] /; m1+m2 != -m3 := TJS[{F1, m1}, {F2, m2}, {F3, m3}] = 0;

(*z-projection is less than total angular momentum*)

TJS[{F1_, m1_}, {F2_, m2_}, {F3_, m3_}] /; Abs[m1] > F1||Abs[m2] > F2||Abs[m3] > F3 := TJS[{F1, m1}, {F2, m2}, {F3, m3}] = 0;

(*TJS returns the value of the Three-J Symbol and stores the value for future use.*)

TJS[{F1_, m1_}, {F2_, m2_}, {F3_, m3_}] := TJS[{F1, m1}, {F2, m2}, {F3, m3}] = ThreeJSymbol[{F1, m1}, {F2, m2}, {F3, m3}];

(*Six-J Symbol*)

(*Triangular conditions [Var88, Sec. 9.1.1, condition (b) ]*)

SJS[{j1_, j2_, j12_}, {j3_, j_, j23_}] /; Abs[j1-j2] > j12||j1+j2 < j12 := SJS[{j1, j2, j12}, {j3, j, j23}] = 0;
SJS[{j1_, j2_, j12_}, {j3_, j_, j23_}] /; Abs[j12-j3] > j||j12+j3 < j := SJS[{j1, j2, j12}, {j3, j, j23}] = 0;
SJS[{j1_, j2_, j12_}, {j3_, j_, j23_}] /; Abs[j2-j3] > j23||j2+j3 < j23 := SJS[{j1, j2, j12}, {j3, j, j23}] = 0;
SJS[{j1_, j2_, j12_}, {j3_, j_, j23_}] /; Abs[j23-j1] > j||j23+j1 < j := SJS[{j1, j2, j12}, {j3, j, j23}] = 0;

(*SJS returns the value of the Six-J Symbol and stores the value for future use.*)

SJS[{j1_, j2_, j12_}, {j3_, j_, j23_}] := SJS[{j1, j2, j12}, {j3, j, j23}] = SixJSymbol[{j1, j2, j12}, {j3, j, j23}];

(*Nine-J Symbol*)

(*NJS returns the value of the Nine-J Symbol function [Var88, Sec. 10.2.4, Eq. (20)]*)

NJS[{a_, b_, c_}, {d_, e_, f_}, {g_, h_, j_}] := NJS[{a, b, c}, {d, e, f}, {g, h, j}] =
    Module[{xmin, xmax},
        xmin = Max[{Abs[a-j], Abs[b-f], Abs[d-h]}];
        xmax = Min[{Abs[a+j], Abs[b+f], Abs[d+h]}];
        Sum[(-1)^(2 x)(2 x+1)SJS[{a, b, c}, {f, j, x}]SJS[{d, e, f}, {b, x, h}]SJS[{g, h, j}, {x, a, d}], {x, xmin, xmax}]];

End[];

EndPackage[];
